package com.ssv.utils;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.ssv.text.ListOperations;
import com.ssv.utils.test.SerializationTest;

/** 
 * Quick implementation (at least much quicker than java.util.Stack) of stack.
 */

public final class ArrayStack<E> implements Stack<E> {

	private static final long serialVersionUID = -7210438026305124449L;

	public ArrayStack() {
		this(DEFAULT_SIZE);
	}

	@SuppressWarnings("unchecked")
	public ArrayStack(int size) {
		stack=(E[])new Object[size];
	}

	/**
	 * This method returns top of the stack without modifying it.
	 */
	public E peek() {
		return stack[stackPointer-1];
	}

	
	/**
	 * Removes top element from the stack, returns is as a result.
	 */
	public E pop() {
		E obj=stack[--stackPointer];
		stack[stackPointer]=null; // To make sure we won't impact GC.
		return obj;
	}

	/**
	 * Insert the element to the top of stack.
	 */
	public E push(E e) {
		return stack[stackPointer++]=e;
	}

	/**
	 * Return number of elements are currently allocated in the stack. 
	 */
	public int size() {
		return stackPointer;
	}	

	/**
	 * Returns true if there is not items in the stack.
	 */
	public boolean empty() {
		return size()==0;
	}

	public Iterator<E> iterator() {
		return new StackIterator();
	}

	public void clear() {
		for(int i=0;i<stackPointer;i++)
			stack[i]=null;
		stackPointer=0;
	}
	
	public ArrayStack<E> clone() {
		return new ArrayStack<E>((E[])stack.clone());	
	}

	public List<E> toList() {
		List<E> copy=new ArrayList<E>();
		for(int i=0;i<stackPointer;i++)
			copy.add(stack[i]);
		return copy;
	}

	public String toString() {
		return "{ "+ListOperations.toString(this)+" }";
	}
	
	@SuppressWarnings("unchecked")
	private void writeObject(java.io.ObjectOutputStream out) throws IOException 
	{
		out.writeInt(stackPointer);
		out.writeInt(stack.length);
		for(int i=0;i<stackPointer;i++)
			out.writeObject(stack[i]);
	}
	
	@SuppressWarnings("unchecked")
	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException
	{
		stackPointer=in.readInt();
		int size=in.readInt();
		stack=(E[])new Object[size];
		for(int i=0;i<stackPointer;i++) 
			stack[i]=(E)in.readObject();
	}
		
	private ArrayStack(E stackcopy[]) {
		stack=stackcopy;
	}
		
	private final class StackIterator implements Iterator<E> {
		public boolean hasNext() {
			return current<stackPointer;
		}

		public E next() {
			return (E)stack[current++];
		}

		public void remove() {
			throw new NoSuchMethodError();
		}
		
		private int current=0;
	}
	
	transient private int stackPointer=0;
	transient private E stack[];

	private static final int DEFAULT_SIZE=50;
	
	@SuppressWarnings("unchecked")
	public static void main(String[] args) {
		Stack<String> ns=new ArrayStack<String>();
		
		ns.push("1");
		ns.push("2");
		ns.push("3");
		ns.push("4");
		System.out.println(ns.pop());
		System.out.println(ns.peek());
		System.out.println(ns.push("5"));
		System.out.println(ns.pop());
		System.out.println(ns.toString());
		SerializationTest.writeObject(ns, ArrayStack.class.getName());
		Stack<String> rs=(Stack<String>)SerializationTest.readObject(ArrayStack.class.getName());
		System.out.println(rs.pop());
		System.out.println(rs.toString());
	}
}
