package com.ssv.utils.parser;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.ssv.utils.ArrayStack;
import com.ssv.utils.Stack;
import com.ssv.utils.parser.Parser.ParserContext;
import com.ssv.utils.parser.Parser.Token;
import com.ssv.utils.tree.BinaryTree;

/**
 * This class is a helper class to parse ariphmetic expression - PN, RPN and 
 * regular infix expressions.
 * 
 * @author Sergey Siryk
 *
 */
public abstract class ExpressionParser<ExpressionLeaf extends BinaryTree<ExpressionLeaf>> {

	static final private int LEFT_TO_RIGHT = 0;
	static final private int RIGHT_TO_LEFT = 1;

	static final private int UNARY = 1;	
	static final private int BINARY = 2;	
	
	static private class Operation {
		int _operands;
		Integer _tokens[];
		int _direction;
		
		public Operation(int operands, Integer tokens[], int direction) {
			_operands = operands;
			_tokens = tokens;
			_direction = direction;
		}
		
		public boolean contains(int token) {
			for(Integer tk:_tokens) {
				if(tk==token)
					return true;
			}
			return false;
		}
	}

	static private class Brackets {
		int _left;
		int _right;
		
		public Brackets(int left, int right) {
			_left = left;
			_right = right;
		}
	}
	
	final private List<Operation> precedence = new ArrayList<Operation>();
	final private Map<Integer, Brackets> brackets = new HashMap<Integer, Brackets>();
	final private Set<Integer> polish_notation_list = new HashSet<Integer>();
	
	public ExpressionParser(ExpressionLeafProvider<ExpressionLeaf> provider) {
		this.provider=provider;
	}
	
	final ExpressionLeafProvider<ExpressionLeaf> provider;
	
	protected void addBinary(Integer... tokens) {
		precedence.add(new Operation(BINARY, tokens, LEFT_TO_RIGHT));
		for(Integer t:tokens)
			polish_notation_list.add(t);
	}
	
	protected void addBinaryRightToLeft(Integer... tokens) {
		precedence.add(new Operation(BINARY, tokens, RIGHT_TO_LEFT));
		for(Integer t:tokens)
			polish_notation_list.add(t);
	}

	protected void addUnary(Integer... tokens) {
		precedence.add(new Operation(UNARY, tokens,LEFT_TO_RIGHT));
	}	

	protected void addOperatorBrackets(int leftBracketToken, int rightBracketToken) {
		brackets.put(leftBracketToken, new Brackets(leftBracketToken,rightBracketToken));
	}
	
	ExpressionLeaf makeTreeRightToLeft(
			final List<ExpressionLeaf> leafs,
			final List<Token> opers) {
		int leafIndex = leafs.size()-1;
		int operIndex = opers.size()-1;
		ExpressionLeaf top =
			provider.makeNode(
				opers.get(operIndex),
				leafs.get(leafIndex-1),
				leafs.get(leafIndex));
		leafIndex-=2;
		operIndex--;
		while(operIndex>=0) {
			top = provider.makeNode(
					opers.get(operIndex--),
					leafs.get(leafIndex--),
					top);
		}
		return top;
	}

	ExpressionLeaf makeTreeLeftToRight(
			final List<ExpressionLeaf> leafs,
			final List<Token> opers) {
		ExpressionLeaf top = provider.makeNode(
				opers.get(0),leafs.get(0),leafs.get(1));
		int leafIndex = 2;
		int operIndex = 1;
		while(operIndex<opers.size()) {
			top = provider.makeNode(
					opers.get(operIndex++),
					top,
					leafs.get(leafIndex++));
		}
		return top;
	}
		
	/**
	 *  Parses the ariphmetic expression, infix notation. 
	 */
	public ExpressionLeaf parse(Parser parser) throws IOException, ParserException {
		ExpressionLeaf leaf =_parse(parser,0); 
		return simplify(leaf);
	}
	
	/**
	 * After execution, we may try to calculate something on "compile time". 
	 * @param parse
	 * @return
	 * @throws ParserException 
	 */
	public ExpressionLeaf simplify(ExpressionLeaf e) throws ParserException {
		return e; 
	}

	public ExpressionLeaf parseRPN(Parser parser) throws IOException, ParserException {
		
		Stack<ExpressionLeaf> stack = new ArrayStack<ExpressionLeaf>();
		while(true) {
			if(isLeaf(parser.token())) {
				stack.push(parseLeaf(parser));
			} else {
				if(! polish_notation_list.contains(parser.token())) {
					break;
				}
				try {
					ExpressionLeaf op2 = stack.pop();
					ExpressionLeaf op1 = stack.pop();
					stack.push(provider.makeNode(parser.ctoken(),op1,op2));
					parser.next();
				} catch(Throwable t) {
					throw new ParserException("Cannot parse RPN:"+t.getMessage(),parser.fetchContext());
				}
			}
		}
		
		if(stack.size()!=1)
			throw new ParserException("Cannot parse RPN, expression terminated unexpectedly.",parser.fetchContext());
		return stack.pop();
	}	

	public ExpressionLeaf parsePN(Parser parser) throws IOException, ParserException {
		if(isLeaf(parser.token())) {
			return parseLeaf(parser);
		}
		if(! polish_notation_list.contains(parser.token())) {
			throw new ParserException("Cannot parse PN, invalid symbol:"+parser.stoken(),parser.fetchContext());
		}		
		Token t = parser.ctoken();
		parser.next();
		return provider.makeNode(t, parsePN(parser),parsePN(parser));
	}	
		
	private ExpressionLeaf _parse(Parser parser, int level) throws IOException, ParserException {
		if(level>=precedence.size()) { // it's a leaf - let's get appropriate parsing.
			Brackets br = brackets.get(parser.token());
			
			if(br!=null) { // Bracketed expression.
				parser.next();
				ExpressionLeaf leaf = _parse(parser, 0);
				parser.accept(br._right);
				return leaf;
			}
			
			return parseLeaf(parser);
		}
		
		Operation op = precedence.get(level);
		
		if(op._operands==UNARY) { // Unary operations.
			Token tk = parser.ctoken();
			if(op.contains(tk.code())) {
				parser.next();
				ExpressionLeaf leaf = _parse(parser, level); 
				return provider.makeNode(
						tk,null,leaf); 
			}
			return _parse(parser,level+1);
		} else {
			final List<ExpressionLeaf> leafs = new ArrayList<ExpressionLeaf>();
			final List<Token> opers = new ArrayList<Token>();
			leafs.add(_parse(parser, level+1));
			while(op.contains(parser.token())) {
				opers.add(parser.ctoken());
				parser.next();
				leafs.add(_parse(parser, level+1));
			}
			if(opers.size()==0)
				return leafs.get(0);
			else {
				if(op._direction==LEFT_TO_RIGHT) {
					return makeTreeLeftToRight(leafs, opers);
				} else {
					return makeTreeRightToLeft(leafs, opers);
				}
			}
		}
	}
	
	abstract public ExpressionLeaf parseLeaf(Parser parser) throws ParserException, IOException;
	abstract public boolean isLeaf(int token) throws ParserException, IOException;
	
	/*
	 * Unit testing.
	 */
	static class MyParser extends Parser {
		public MyParser() throws IOException,
				ParserException {
			super(false);
			addSpace(' ');
			addSpace('\t');
			addSpace('\r');
			addSpace('\n');
			addCommentSymbols("/*", "*/");
			addLineCommentSymbol("//");
			addToken(LBR,"(");
			addToken(RBR,")");
			addToken(PLUS,"+");
			addToken(MINUS,"-");
			addToken(DIV,"/");
			addToken(MUL,"*");
			addToken(EQ,"==");
			addToken(GT,">");
			addToken(GE,">=");
			addToken(LT,"<");
			addToken(LE,"<=");
			addToken(ASSIGN,"=");
		}
		final static int LBR = 1;
		final static int RBR = 2;
		final static int PLUS = 3;
		final static int MINUS = 4;
		final static int DIV = 5;
		final static int MUL = 6;
		final static int EQ = 7;
		final static int GT = 8;
		final static int GE = 9;
		final static int LT = 10;
		final static int LE = 11;
		final static int ASSIGN = 12;
	}	

	static public class IdentExpressionLeaf extends BinaryTree<IdentExpressionLeaf> {
		public IdentExpressionLeaf(Token token, 
				IdentExpressionLeaf left,
				IdentExpressionLeaf right) {
			super(left, right);
			_ident = null;
			this.token = token;
		}

		public IdentExpressionLeaf(String ident) {
			super(null,null);
			_ident = ident;
			this.token = null;
		}

		public String toString() {
			return (_ident!=null)?_ident.toString():token.toString();
		}
		
		private static final long serialVersionUID = -3064327124242652083L;
		final Token token;
		final String _ident;
	}

	static class MyLeafProvider implements ExpressionLeafProvider<IdentExpressionLeaf> {
		public IdentExpressionLeaf makeNode(Token t, IdentExpressionLeaf left,
				IdentExpressionLeaf right) {
			return new IdentExpressionLeaf(t,left,right);
		}
	}
	
	static class MyExprParser extends ExpressionParser<IdentExpressionLeaf> {		
		public MyExprParser() {
			super(new MyLeafProvider());
			addOperatorBrackets(MyParser.LBR,MyParser.RBR);
			addBinaryRightToLeft(MyParser.ASSIGN);
			addBinary(MyParser.EQ,MyParser.GT,MyParser.GE,MyParser.LT,MyParser.LE);
			addBinaryRightToLeft(MyParser.PLUS,MyParser.MINUS);
			addBinary(MyParser.DIV,MyParser.MUL);
		}
		
		public ExpressionParser.IdentExpressionLeaf parseLeaf(
				Parser parser) throws ParserException, IOException {
			ParserContext ctx = parser.fetchContext();
			if(parser.token()==Parser.IDENTIFIER) {
				IdentExpressionLeaf leaf = new IdentExpressionLeaf(parser.getIdentifier());
				parser.next();
				return leaf;
			}
				
			throw new ParserException("Invalid symbol:"+parser.stoken(),ctx);
		}

		public boolean isLeaf(int token) throws ParserException, IOException {
			return token==Parser.IDENTIFIER;
		}
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			MyParser tokens = new MyParser();
			tokens.init(System.in);
			MyExprParser exprParser = new MyExprParser();
			exprParser.parse(tokens).print(System.out);
//			exprParser.parseRPN(tokens).print(System.out);
//			exprParser.parsePN(tokens).print(System.out);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
