package com.ssv.utils.parser;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.ssv.utils.tree.MapTree;

/**
 * This class is substitution for Java's StringTokenizer. Actually, this is much
 * more powerful and flexible tokenizer, it allow user to specify his own
 * keyword, symbols, the way of doing commentary etc. There are 6 predefined
 * tokens:
 * <p>
 * <b>EOF</b> - end of file reached. <br>
 * <b>IDENTIFIER</b> - identifier encountered, it's name accessible through
 * getIdentifier() <br>
 * <b>NUMBER</b> -number encountered, value accessible by getNumber() <br>
 * <b>DOUBLENUMBER</b> - float number encountered, value accessible by
 * getDoubleNumber() <br>
 * <b>SINGLEQUOTEDSTRING</b> single quoted string literal encountered, value
 * accessible by getString() <br>
 * <b>DOUBLEQUOTEDSTRING</b> double quoted string literal encountered, value
 * accessible by getString()<br>
 * <p>
 * Also, the user can define his own symbols (see below).
 * <p>
 * To define keyword use {@link #addKeyword(String)}. Store the returned int
 * value to recognize the token when it encounters in the input stream. The
 * example is here:
 * <p>
 * 
 * <pre><code>
 *                  		final static int BEGIN = addKeyword(&quot;BEGIN&quot;);
 *                  		...
 *                  		if(token()==BEGIN) { // BEGIN just scanned from the input stream.
 *                  			...
 *                  		}
 * </code></pre>
 * 
 * <p>
 * The same way should be used when declare your own symbols with <b>addToken</b>
 * method. There are 2 versions of this method - {@link #addToken(String)} and
 * {@link #addToken(String, String)}. First one just declares new symbol and it
 * will be shown in diag messages exactly like it was declared (which is OK when
 * symbol doesn't have any non-printing symbols). If symbol incudes such
 * symbols, it is more practical to assign string name for it. In this example
 * we declare LF token (i.e. end of the line)
 * <p>
 * 
 * <pre><code>
 *     final static int LF = addToken(&quot;\r\n&quot;, &quot;LF&quot;);
 *     ...
 *     if(token()==LF) { 
 *     	  // Printing log message, actually &quot;LF encountered.&quot; will be printed.
 *     	  System.out.println(tokenToStr(token())+&quot; encountered.&quot;);
 *     }
 * </code></pre>
 * 
 * <p>
 * 
 * Also there is a way to handle comments, namely
 * {@link #addCommentSymbols(String, String)} and
 * {@link #addLineCommentSymbol(String)}. First of them registers the way to
 * mark something from "begin" to "end" as a comments. For example, if you do
 * this:
 * 
 * <p>
 * 
 * <pre><code>
 * addCommentSymbols(&quot;{&quot;, &quot;}&quot;);
 * </code></pre>
 * 
 * <p>
 * 
 * you will have Pascal's style of comments, which mean everything between '{'
 * and '}' will be skipped by parser.<br>
 * Another comments style can be handled by
 * {@link #addLineCommentSymbol(String)}. The difference is here - you should
 * indicate begin of commentary and everything after that will be skipped until
 * line ends.<br>
 * If you want to analyze contents of commentary, you may set
 * {@link #setCommentHandler(CommentHandler)} listener (see below).
 * <p>
 * One more important matter here is how to handle whitespaces. You may set as
 * many whitespace symbols by invoking {@link #addSpace(char)}. Usually we
 * treat spacebars, tabs and end of lines as whitespaces, so probably you will
 * have to do this:
 * 
 * <p>
 * 
 * <pre><code>
 * addSpace(' ');
 * addSpace('\r');
 * addSpace('\n');
 * addSpace('\t');
 * </code></pre>
 * 
 * <p>
 * 
 * But it is not obligatory action, for example you may handle end of line as a
 * meaningful token.
 * 
 * @author Sergey Siryk <b>sergey.siryk@gmail.com</b>
 */
public class Parser {

	private final static class Identifier {
		private String s;
		public Identifier(String s) {
			this.s=s;
		}
		public String toString() {
			return s;
		}
	}
	
	public final static class ParserContext {
		ParserContext(int col, int row) {
			o=null;
			this.col=col;
			this.row=row;
		}

		ParserContext(ParserContext src, Identifier ident) {
			this(src,(Object)ident);
		}

		ParserContext(ParserContext src, String str) {
			this(src,(Object)str);
		}

		ParserContext(ParserContext src, long num) {
			this(src,new Long(num));
		}
		
		ParserContext(ParserContext src, double doublenum) {
			this(src, new Double(doublenum));
		}
		
		public ParserContext clone() {
			return new ParserContext(this,o);
		}
		
		private ParserContext(ParserContext src, Object obj) {
			o=obj;
			this.col=src.getColumn();
			this.row=src.getRow();
		}
				
		public String getIdentifier() {
			if(o.getClass()!=Identifier.class)
				throw new Error("Context is not identifier.");
			return ((Identifier)o).toString(); 
		}

		public String getString() {
			if(o.getClass()!=String.class)
				throw new Error("Context is not string.");
			return (String)o; 
		}

		public long getLongNumber() {
			if(o.getClass()!=Long.class)
				throw new Error("Context is not long number.");
			return ((Long)o).longValue(); 
		}

		public double getDoubleNumber() {
			if(o.getClass()!=Double.class)
				throw new Error("Context is not double number.");
			return ((Double)o).doubleValue(); 
		}

		public int getColumn() { return col; }
		public int getRow() { return row; }

		private Object o;
		private int col, row;
	}
	
	/**
	 * GeneralTokenStream constructor.
	 * 
	 * @param is
	 *            input stream to read from.
	 * @param cs
	 *            true if our grammar is case-sensitive, false otherwise.
	 * @throws ParserException 
	 */
	protected Parser(boolean cs) throws ParserException {

		addBasicToken(_NULL_, "*NULL*");
		addBasicToken(EOF, "EOF");
		addBasicToken(IDENTIFIER, "IDENTIFIER");
		addBasicToken(NUMBER, "NUMBER");
		addBasicToken(DOUBLENUMBER, "DOUBLENUMBER");
		addBasicToken(SINGLEQUOTEDSTRING, "SINGLEQUOTEDSTRING");
		addBasicToken(DOUBLEQUOTEDSTRING, "DOUBLEQUOTEDSTRING");
		addBasicToken(NONNUMBER, "NONNUMBER");

		caseSensitive = cs;
	}

	/**
	 * Method opens the stream to read from. If another stream has been opened
	 * previously, it will be closed.
	 * 
	 * @param is stream to open.
	 */
	public void init(InputStream is) throws IOException {
		if(input!=null)
			input.close();
		input = new InputStreamReader(is);
		initialized=false;
		token=IDENTIFIER;
	}
	
	private String listOfTokens(Integer tokens[]) {
		StringBuffer sb=new StringBuffer();
		boolean first=true;
		for(Integer t:tokens) {
			if(! first)
				sb.append(", ");
			first=false;
			sb.append(stokenshort(t));
		}
		return sb.toString();
	}
	/**
	 * Checks if current token is what we expect, reads next token. Using this
	 * method could be tricky since you going to lose current token. For example,
	 * 
	 * <pre><code>
	 *		parser.accept(IDENTIFIER);
	 *		System.out.println(parser.getIdentifier());
	 * </code></pre>
	 * 
	 * will not work since value of identifier is already lost by time we try to
	 * get it. This code should be done this way:
	 * 
	 * <pre><code>
	 * 		String id_value=parser.getIdentifier();
	 *		parser.accept(IDENTIFIER);
	 *		System.out.println(id_value);
	 * </code></pre>
	 * 
	 * But it still not the best way to do it. There is special function which
	 * holds context of accepted symbol: accepted(). You may use it like that:
	 *  
	 * <pre><code>
	 *		parser.accept(IDENTIFIER);
	 *		System.out.println(parser.accepted().getIdentifier());
	 * </code></pre>
	 * 
	 * @param tokens -
	 *            tokens we expect.
	 * @throws ParserException
	 *             if current token is not what we expect.
	 * @throws IOException
	 *             if there is input stream problem.
	 */
	public int accept(Integer... tokens) throws ParserException, IOException {
		int rez=-1;
		for(Integer tok:tokens) {
			if(tok==token()) {
				rez=tok;
				break;
			}
		}
		if (rez==-1) {
			error("Invalid symbol:" + stoken() + ", expecting:"
					+ listOfTokens(tokens));
		}
		acceptcontext=fetchContext();
		next();
		return rez;
	}

	/**
	 * Creates copy of current context.
	 */
	public ParserContext fetchContext() {
		return context.clone();
	}

	/**
	 * Return link to context which has been 
	 * created inside last accept() method.
	 * 
	 * See also {@link #accept()}.
	 */
	public ParserContext accepted() {
		return acceptcontext;
	}
	
	/**
	 * @return value of last parsed identifier.
	 */
	public String getIdentifier() {
		return context.getIdentifier();
	}

	/**
	 * @return value of last parsed integer constant.
	 */
	public long getNumber() {
		return context.getLongNumber();
	}

	/**
	 * @return value of last parsed float constant.
	 */
	public double getDoubleNumber() {
		return context.getDoubleNumber();
	}

	/**
	 * @return value of last parsed string literal.
	 */
	public String getString() {
		return context.getString();
	}

	/**
	 * Reads next token from the input stream.
	 * 
	 * @return last parsed token.
	 * @throws <b>IOException
	 *             </b> if there is an error while reading from the input
	 *             stream.
	 * @throws <b>GeneralTokenException
	 *             </b> syntax error.
	 */
	public int next() throws IOException, ParserException {
		if (!initialized) {
			getCh();
			initialized=true;
		}
		if(token==EOF)
			return EOF;
		while (true) {
			token = this.nextPreliminarySymbol();
			if (isLineComment(token)) {
				skipToLineEnd();
				continue;
			}
			int commentId = isComment(token);
			if (commentId != -1) {
				skipTo(commentEnds.get(commentId));
				continue;
			}
			break;
		}
		if (token == IDENTIFIER) {
			Integer tk;
			if (caseSensitive)
				tk = keywords.get(getIdentifier());
			else
				tk = keywords.get(getIdentifier().toLowerCase());
			if (tk != null)
				return token = tk.intValue();
		}
		return token;
	}

	/**
	 * Returns current symbol.
	 * @throws ParserException 
	 * @throws IOException 
	 */
	public int token() throws IOException, ParserException {
		if (!initialized) {
			next();
		}
		return token;
	}

	public final static class Token {
		public Token(int token, String info, ParserContext ctx) {
			this.token=token;
			this.ctx=ctx;
			this.info=info;
		}

		public int code() {
			return token;
		}

		public ParserContext ctx() {
			return ctx;
		}		

		public String info() {
			return info;
		}		
		
		public String toString() {
			return info;
		}
		
		private final int token;
		private final ParserContext ctx;
		private String info;
	}
	
	/**
	 * Returns current symbol.
	 * @throws ParserException 
	 * @throws IOException 
	 */
	public Token ctoken() throws IOException, ParserException {
		if (!initialized) {
			next();
		}
		return new Token(token, stoken(), fetchContext());
	}	
	
	/**
	 * Sets listener to invoke when commentary encountered. Whole commentary's
	 * body will be passed to the listener if it is defined.
	 * 
	 * @param ch
	 */
	public void setCommentHandler(CommentHandler ch) {
		commentHandler = ch;
	}

	/**
	 * Returns string representation of the current token. Be careful - current
	 * implementation is not accurate, it returns string like this:<br>
	 * IDENTIFIER [ identname ]<br>
	 * where identname is <b>current</b> value of getIdentifier().
	 * 
	 * @param token
	 * @return
	 */
	public String stoken() {
		String tok = strtokens.get(token);
		String add = "";
		if (token == IDENTIFIER) {
			add = " [ " + getIdentifier() + " ]";
		} else if (token == NUMBER) {
			add = " [ " + getNumber() + " ]";
		} else if (token == DOUBLENUMBER) {
			add = " [ " + getDoubleNumber() + " ]";
		} else if (token == DOUBLEQUOTEDSTRING) {
			add = " [ \"" + getString() + "\" ]";
		} else if (token == SINGLEQUOTEDSTRING) {
			add = " [ \'" + getString() + "\' ]";
		} else if (token == NONNUMBER) {
			add = " [ \'" + getString() + "\' ]";
		}
		return tok + add;
	}

	/**
	 * @param token ID.
	 * @return token's name.
	 */
	public String stokenshort(int ID) {
		return strtokens.get(ID);
	}
		
	/**
	 * Adds commentary definition. Parser will skip everything between values of
	 * <code>begin</code> and <code>end</code> in the input stream.
	 * 
	 * @param begin
	 * @param end
	 * @throws ParserException 
	 */
	protected void addCommentSymbols(String begin, String end) throws ParserException {
		comments.add(new Integer(addToken(symbolCounter++,begin)));
		commentEnds.add(end);
	}

	/**
	 * Adds line commentary definition. Everything between value of
	 * <code>begin</code> and end of the line will be skipped by parser.
	 * 
	 * @param begin
	 * @throws ParserException 
	 */
	protected void addLineCommentSymbol(String begin) throws ParserException {
		linecomments.add(new Integer(addToken(symbolCounter++, begin)));
	}

	/**
	 * Defines new keyword on keywordID.
	 * 
	 * @param keyword
	 * @return token ID
	 * @throws ParserException 
	 */
	protected int addKeyword(int keywordID, String keyword) throws ParserException {
		registerToken(keywordID, keyword);
		String key = keyword;
		if (!caseSensitive)
			key = keyword.toLowerCase();
		keywords.put(key, keywordID);
		return keywordID;
	}

	/**
	 * Registers new token.
	 * 
	 * @param tokenValue -
	 *            symbol to register.
	 * @return int value reserved for this token.
	 * @throws ParserException 
	 */
	protected int addToken(int tokenID, String tokenValue) throws ParserException {
		return addToken(tokenID, tokenValue, tokenValue);
	}

	/**
	 * Registers new token under specified name. The difference from previous
	 * method is this: name of the token, which will be returned by
	 * toString(token) will be differen from token itself. Useful when token
	 * contains non-printing symbols.
	 * 
	 * @param tokenValue -
	 *            token to register.
	 * @param tokenName -
	 *            token's name.
	 * @return reserved token id for this token.
	 * @throws ParserException 
	 */
	protected int addToken(int tokenID, String tokenValue, String tokenName) throws ParserException {
		Node current = root;
		for (int i = 0; i < tokenValue.length(); i++) {
			String value = "" + tokenValue.charAt(i);
			Node subnode = (Node) current.locate(value);
			if (subnode == null) { // There is no this value yet.
				subnode = new Node(value);
				current.add(subnode);
			}
			current = subnode;
		}
		current.token = tokenID;
		registerToken(tokenID, tokenName);
		return current.token;
	}

	/**
	 * Mark symbol as a whitespace.
	 * 
	 * @param ch
	 */
	protected void addSpace(char ch) {
		spaces.add(ch);
	}

	/** 
	 * Token to indicate the token is not assigned yet.
	 */
	static final private int _NULL_ = 11000;

	/**
	 * End of file.
	 */
	static final public int EOF = 11001;

	/**
	 * Identifier. Invoke {@link #getIdentifier()} to access identifier's name.
	 */
	static final public int IDENTIFIER = 11002;

	/**
	 * Long number. Invoke {@link #getNumber()} to get the value.
	 */
	static final public int NUMBER = 11003;

	/**
	 * Float number. Invoke {@link #getDoubleNumber()} to get the value.
	 */
	static final public int DOUBLENUMBER = 11004;
	
	/**
	 * Single-quoted string. Invoke {@link #getString()} to get the value.
	 */
	static final public int SINGLEQUOTEDSTRING = 11005;

	/**
	 * Double-quoted string. Invoke {@link #getString()} to get the value.
	 */
	static final public int DOUBLEQUOTEDSTRING = 11006;
	
	/**
	 * Invalid token, starts with number and. Invoke {@link #getString()} to get the value.
	 */
	static final public int NONNUMBER = 11007;	
	
	/**
	 * Registers "basic" token.
	 * @throws ParserException 
	 */
	private int addBasicToken(int tokenID, String tokenValue) throws ParserException {
		registerToken(tokenID, tokenValue);
		return tokenID;
	}

	private int symbolCounter = 10000;
	private Map<String, Integer> keywords = new HashMap<String, Integer>();
	private Map<Integer, String> strtokens = new HashMap<Integer, String>();
	private Node root = new Node("");
	private HashSet<Character> spaces = new HashSet<Character>();
	private List<Integer> comments = new ArrayList<Integer>();
	private List<String> commentEnds = new ArrayList<String>();
	private Set<Integer> linecomments = new HashSet<Integer>();

	private ParserContext acceptcontext=null;
	private ParserContext context;
		
	private static final char SINGLEQUOT = '\'';
	private static final char DOUBLEQUOT = '\"';
	private static final char LF = '\n';
	private static final char CR = '\r'; // CR+LF
	private static final char BACKSLASH = '\\';

	private final boolean caseSensitive;

	private Reader input;

	private int token;
	private int ch;
	private int line = 0;
	private int pos = 0;

	private CommentHandler commentHandler = null;

	private boolean initialized = false;
	
	static final HashMap<Character, Character> specialSymbols = new HashMap<Character, Character>();
	static {
		specialSymbols.put(new Character('n'), new Character('\n'));
		specialSymbols.put(new Character('r'), new Character('\r'));
		specialSymbols.put(new Character('\\'), new Character('\\'));
		specialSymbols.put(new Character('t'), new Character('\t'));
		specialSymbols.put(new Character('\''), new Character('\''));
		specialSymbols.put(new Character('\"'), new Character('\"'));
	}

	private void registerToken(Integer token, String str) throws ParserException {
		if(strtokens.get(token)!=null)
			error("Token ID:"+token+" is already reserved.");
		strtokens.put(token, str);
	}

	private class Node extends MapTree<String,Node> {
		private static final long serialVersionUID = 3791525231198602718L;

		public Node(String name) {
			super(name);
		}

		public String toString() {
			return super.toString() + ":" + token;
		}
		int token = _NULL_;
	}

	/**
	 * Reads next symbol from the input stream.
	 * @throws ParserException 
	 */
	private int getCh() throws IOException, ParserException {
		if(input==null)
			error("Input stream is NULL. Use init(InputStream is) to initialize stream.");
		ch = input.read();
		if (ch == '\t') {
			pos += 7;
		}
		pos++;
		if (ch == '\r') {
			pos = 0;
		}
		if (ch == '\n') {
			pos = 0;
			line++;
		}
		return ch;
	}

	/**
	 * Skips whitespaces.
	 * @throws ParserException 
	 */
	private void skipSpaces() throws IOException, ParserException {
		while (spaces.contains(new Character(ch())))
			getCh();
	}

	/**
	 * Throws a GeneralTokenException.
	 */
	protected void error(String message) throws ParserException {
		throw new ParserException(message, context.getColumn(), context.getRow());
	}

	protected void error(Exception cause, String message)
			throws ParserException {
		throw new ParserException(cause, message, pos, line);
	}

	private DigitSequence readDigitSequence() throws IOException, ParserException {
		StringBuffer num = new StringBuffer();
		if(Character.isDigit(ch()))
			do {
				num.append(ch());
			} while (Character.isDigit(getCh()));
		return new DigitSequence(num.toString());
	}

	static final private char CPOINT = '.'; 
	static final private char CMINUS = '-'; 
	static final private char CE = 'E'; 
	static final private char CB = 'B'; 
	
	private int readNumber() throws IOException, ParserException {
		DigitSequence num = readDigitSequence();
		
		DigitSequence fraq = null;

		if(ch()==CPOINT) {
			getCh();
			fraq = readDigitSequence();
		}
		
		boolean negativeE = false;
		DigitSequence E = null;
		
		if(ch()==CE) {
			if(getCh()==CMINUS) {
				negativeE=true;
				getCh();
			}
			E=readDigitSequence();
		}

		DigitSequence B = null;

		if(ch()==CB) {
			getCh();
			B=readDigitSequence();
		}				

		StringBuffer NaN = null;
		
		if(Character.isJavaIdentifierPart(ch())) { // Error in number.
			NaN = new StringBuffer();
			do {
				NaN.append(ch());
			} while(Character.isJavaIdentifierPart(getCh()));
		}

		boolean success = NaN==null;
		if(B!=null) {
			long b = B.toLong(10);
			if(b<2 || b>10)
				success = false;
			if(num.maxDigit()>=b)
				success=false;
			if(fraq!=null && fraq.maxDigit()>=b)
				success=false;
		}		
		
		if(! success) {
			StringBuffer all = new StringBuffer();
			all.append(num);
			if(fraq!=null)
				all.append(""+CPOINT+fraq);
			if(E!=null) {
				all.append(""+ CE+((negativeE)?"-":"")+E);
			}
			if(B!=null) {
				all.append(""+CB+B);
			}
			if(NaN!=null)
				all.append(NaN.toString());
			context=new ParserContext(context, all.toString());
			return NONNUMBER;
		}

		if(fraq==null && (E==null || ! negativeE)) {
			context=new ParserContext(context, makeLong(num,negativeE,E,B));
			return NUMBER;					
		}

		context=new ParserContext(context, makeDouble(num,fraq,negativeE,E,B));
		return DOUBLENUMBER;						
	}
	
	private Double makeDouble(DigitSequence num, DigitSequence fraq,
			boolean negativeE,
			DigitSequence e, DigitSequence b) {
		int base = (int)((b==null)?10:b.toLong(10));
		double rez = num.toLong(base);
		double fr = (fraq==null)?0:fraq.toLong(base);
		int sz = (fraq==null)?0:fraq.size();
		rez=rez+fr*Math.pow(base, -sz);
		if(e!=null) {
			long le = e.toLong(10);
			if(negativeE)
				le=-le;
			rez*=Math.pow(base, le);
		}
		return rez;
	}

	private Long makeLong(DigitSequence num,
			boolean negativeE,
			DigitSequence e,			
			DigitSequence b) {
		int base = (int)((b==null)?10:b.toLong(10));
		long rez = num.toLong(base);
		if(e!=null)
			rez*=Math.pow(base, e.toLong(10));
		return rez;
	}

	private int nextPreliminarySymbol() throws IOException,
			ParserException {
		while (true) {
			skipSpaces();
			context=new ParserContext(pos,line);
			Node found = (Node) root.locate("" + ch());
			if (found != null) { // This is not identifier or constant
				do {
					getCh();
					Node c = (Node) found.locate("" + ch());
					if (c == null)
						break;
					found = c;
				} while (true);
				if (found.token == _NULL_)
					error("Symbol is invalid:" + ch());
				return found.token;
			} else if (Character.isJavaIdentifierStart(ch())) { // IDENTIFIER
				StringBuffer sb = new StringBuffer();
				sb.append(ch());
				while (Character.isJavaIdentifierPart((int) getCh())) {
					sb.append(ch());
				}
				context=new ParserContext(context,new Identifier(sb.toString()));
				return IDENTIFIER;
			} else if (ch() == SINGLEQUOT || ch() == DOUBLEQUOT) { // It's
				// single
				// quoted
				// number.
				char end = ch();
				StringBuffer sb = new StringBuffer();
				getCh();
				while (ch() != end && ch() != LF && ch() != CR) {
					char sym = ch();
					if (ch() == BACKSLASH) {
						Character special = specialSymbols.get(new Character(
								(char) getCh()));
						if (special == null)
							error("Special symbol \'" + ch()
									+ "\' does not exist.");
						sym = special.charValue();
					}
					sb.append(sym);
					getCh();
				}
				if (ch() != end)
					error("String literal is not closed.");
				context=new ParserContext(context, sb.toString());
				getCh();
				return (end == SINGLEQUOT) ? SINGLEQUOTEDSTRING
						: DOUBLEQUOTEDSTRING;
			} else if (Character.isDigit(ch())) {
				return readNumber();
			} else if (ich() == -1) {
				return EOF;
			}
			error("Invalid symbol: \'" + ch() + "\'");
		}
	}

	private int isComment(int token) {
		for (int i = 0; i < comments.size(); i++)
			if (comments.get(i).intValue() == token)
				return i;
		return -1;
	}

	private boolean isLineComment(int token) {
		return linecomments.contains(new Integer(token));
	}

	private void handleComment(String comment) throws ParserException,
			IOException {
		if (commentHandler != null) {
			try {
				commentHandler.processComment(comment);
			} catch (Exception e) {
				error(e, "Error while processing commentary.");
			}
			while (ch() == CR || ch() == LF) {
				getCh();
			}
		}
	}

	private void skipToLineEnd() throws IOException, ParserException {
		StringBuffer sb = new StringBuffer();
		while (ich() != -1 && ch() != CR && ch() != LF) {
			sb.append(ch());
			getCh();
		}
		handleComment(sb.toString());
	}

	private void skipTo(String pattern) throws IOException,
			ParserException {
		StringBuffer sb = new StringBuffer();
		while (ich() != -1) {
			int currentPos = 0;
			while (ch() == pattern.charAt(currentPos)) {
				getCh();
				if (++currentPos >= pattern.length()) { // Got end.
					String res=sb.toString();
					handleComment(res.substring(0,res.length()-pattern.length()));
					return;
				}
			}
			sb.append(pattern.substring(0, currentPos));
			getCh();
			sb.append(ch());
		}
		error("Unexpected end of file.");
	}

	private char ch() {
		return (char) ch;
	}

	private int ich() {
		return (int) ch;
	}

	/*
	 * Unit testing.
	 */
	static class MyParser extends Parser {
		public MyParser() throws IOException,
				ParserException {
			super(false);
			addSpace(' ');
			addSpace('\t');
			addSpace('\r');
			addSpace('\n');
			addCommentSymbols("/*", "*/");
			addLineCommentSymbol("//");
			addToken(SEMICOLON,";");
			addToken(DIV,"/");
			addToken(DIVE,"/=");
			addKeyword(CREATE,"CREATE");
		}
		final static int SEMICOLON = 1;
		final static int DIV = 2;
		final static int DIVE = 3;
		final static int CREATE = 4;
	}

	public static void main(String[] args) {
		try {
			MyParser tokens = new MyParser();
			tokens.init(System.in);
			while (tokens.token() != EOF) {
				System.out.println(tokens.stoken());
				tokens.next();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
