package com.ssv.utils.tree;

import java.io.PrintStream;
import java.util.Iterator;

import com.ssv.utils.test.SerializationTest;

/**
 * Basic implementation of binary tree. User has to override at least toString 
 * method to use it.  
 * @author Sergey Siryk
 *
 * @param <TreeNode> - type of leafs used in this tree.
 */
public class BinaryTree<Leaf extends BinaryTree<Leaf>> 
extends Tree<Leaf> {
	private static final long serialVersionUID = 4490025179981716799L;

	/**
	 * This constructor creates isolated leaf with no childs attached.
	 *
	 */
	public BinaryTree() {
		super();
	}

	
	/**
	 * This constructor creates a tree with subtrees. If both <code>left</code> and 
	 * <code>right</code> are <code>null</code>, this is exactly same as parameterless version.
	 * @param left 
	 * @param right
	 */
	public BinaryTree(Leaf left,Leaf right) {
		this();
		setLeft(left);
		setRight(right);
	}	
	
	/**
	 * Goes through the tree in the infix order (left branch, root, right branch).
	 * It could be used for listing all the items or to search some specific item. 
	 * To perform such search, <code>action</code> should return non-null value
	 * when it is invoked on the node it looks for.
	 * 
	 * @param action - action to perform on each node.
	 * @return null if nothing is found/action is not "searching", reference to found node if 
	 * something is found.
	 * @see com.ssv.util.ITreeAction
	 */
	
	final public Leaf infix(ITreeAction<Leaf> action) {
		BinaryTree<Leaf> tree;
		Leaf rez;
		
		action.enter();
		tree=getLeft();
		if (tree != null) { 
			if ((rez = tree.infix(action)) != null)
				return rez;
		} 
		action.exit();

		tree=this;
		if ((rez = action.performAction(tree)) != null)
			return rez;

		action.enter();
		tree=getRight();
		if (tree != null) { 
			if ((rez = tree.infix(action)) != null)
				return rez;
		} 
		action.exit();
		return null;
	}

	/**
	 * Goes through the tree in the inverted infix order (right branch, root, left branch).
	 * It could be used for listing all the items or to search some specific item. 
	 * To perform such search, <code>action</code> should return non-null value
	 * when it is invoked on the node it looks for.
	 * 
	 * @param action - action to perform on each node.
	 * @return null if nothing is found/action is not "searching", reference to found node if 
	 * something is found.
	 * @see com.ssv.util.ITreeAction
	 */
	
	final public Leaf infixInverted(ITreeAction<Leaf> action) {
		BinaryTree<Leaf> tree;
		Leaf rez;
		
		action.enter();
		tree=getRight();
		if (tree != null) { 
			if ((rez = tree.infixInverted(action)) != null)
				return rez;
		} 
		action.exit();

		tree=this;
		if ((rez = action.performAction(tree)) != null)
			return rez;

		action.enter();
		tree=getLeft();
		if (tree != null) { 
			if ((rez = tree.infixInverted(action)) != null)
				return rez;
		} 
		action.exit();
		return null;
	}	
	
	/**
	 *  Prints all the nodes to the <code>ps</code> in infix order.
	 */
	
	final public void print(PrintStream ps) {
		infixInverted(new PrintOperation<Leaf>(ps));
	}

	/** 
	 * Attaches the give subtree to the node. It implemented only for compatibility
	 * with basic <em>Tree</em>. Normally, this code shouldn't be used, binary tree
	 * should use <code>setLeft</code> and <code>setRight</code> instead.
	 */
	
	final protected boolean addNode(Leaf tree) {
		if(left==null) {
			left=tree;
		}
		else if(right==null) {
			right=tree;
		} else 
			throw new AssertionError("Binary item cannot have more than 2 subnodes."); 		
		return true;
	}

	final protected boolean removeNode(Leaf tree) {
		boolean changed=false;
		if(tree==getLeft()) {
			setLeft(null);
			changed=true;
		}
		if(tree==getRight()) {
			setRight(null);
			changed=true;
		}
		return changed;
	}
	
	final public int size() {
		int sz=0;
		if(left!=null)
			sz++;
		if(right!=null)
			sz++;
		return sz;
	}

	final public Leaf getNode(int index) {
		if(index>=size())
			throw INVALID_INDEX;
		int c=0;
		if(left!=null) {
			if(index==c++)
				return left;
		}
		if(right!=null) {
			if(index==c)
				return right;
		}
		throw INVALID_INDEX;
	}
		
	/**
	 * Sets new left subnode.
	 * 
	 * @param subtree - new subnode.
	 */
	final public void setLeft(Leaf subtree) {
		if(left==subtree)
			return;
		if(left!=null)
			left.setParent(null);
		left=subtree;
		if(subtree!=null)
			subtree.setParent(this);
	}

	/**
	 * Sets new right subnode.
	 * 
	 * @param subtree - new subnode.
	 */
	final public void setRight(Leaf subtree) {
		if(right==subtree)
			return;
		if(right!=null)
			right.setParent(null);
		right=subtree;
		if(subtree!=null)
			subtree.setParent(this);
	}
		
	/**
	 * @return left subnode of the current node.
	 */
	final public Leaf getLeft() {
		return left;
	}

	/**
	 * @return right subnode of the current node.
	 */
	final public Leaf getRight() {
		return right;
	}		
	
	transient private Leaf left,right;
	transient static private Error INVALID_INDEX = new AssertionError("Index is out of bounds.");
	
	/* UNIT TEST */
	
	@SuppressWarnings("unchecked")
	public static void main(String args[]) {
		class DT extends BinaryTree<DT> {
			private static final long serialVersionUID = 1L;

			public DT(String name) {
				super();
				this.name = name;
			}

			public DT(String name, DT left, DT right) {
				super(left,right);
				this.name = name;
			}			
			
			public String toString() {
				return name;
			}

			String name;
		}

		DT root = new DT("root",
					 new DT("1",
							 new DT("11"),
							 new DT("12",new DT("121"),null)),
					 new DT("2")		
		);

		root.getRight().add(new DT("21"));
		
		System.out.println("*** Prefix:");
		root.printPrefix(System.out);
		System.out.println("*** Infix:");
		root.print(System.out);
		System.out.println("***");
		for(Iterator<DT> i=root.iterator();i.hasNext();) {
			System.out.println(i.next());
		}
		System.out.println("*** Tree:");
		Tree<DT> tree=root;
		tree.printPostfix(System.out);
		
		SerializationTest.writeObject(tree, BinaryTree.class.getName());
		Tree t = (Tree) SerializationTest.readObject(BinaryTree.class.getName());
		t.print(System.out);
	}
}
