package com.ssv.utils.tree;

import java.util.Iterator;

import com.ssv.utils.BuiltinMap;
import com.ssv.utils.CompoundName;
import com.ssv.utils.StringCompoundName;
import com.ssv.utils.test.SerializationTest;

public class MapTree<K, Leaf extends MapTree<K,Leaf>> extends 
Tree<Leaf> implements BuiltinMap.Value<K> {
	private static final long serialVersionUID = -456185147067805248L;

	public MapTree(K key) {
		super();
		this.key = key;
	}

	public String toString() {
		return key.toString();
	}

	final protected boolean addNode(Leaf leaf) {
		if (map().contains(leaf))
			return false;
		map().put(leaf);
		return true;
	}
		
	final protected boolean removeNode(Leaf leaf) {
		if (!map().containsKey(leaf))
			return false;
		map().remove(leaf);
		return true;
	}

	final public int size() {
		return map().size();
	}

	final public K getKey() {
		return key;
	}

	public MapTree<K,Leaf> locate(CompoundName<K> dottedName, boolean rootIsShown) {
		int index=1;
		if(rootIsShown) {
			if(dottedName.numTokens()==0)
				throw new AssertionError("Name is invalid, cannot search by empty name.");
			if(! dottedName.getToken(0).toString().equals(toString())) {
				return null;
			} 
		} else {
			if(dottedName.numTokens()==0)
				return this;
			index=0;
		}
		MapTree<K,Leaf> current=this;
		while(index<dottedName.numTokens()) {
			MapTree<K,Leaf> c=current.locate(dottedName.getToken(index++));
			if(c==null)
				return null;
			current=c;
		}
		return current;
	}	
			
	/**
	 * The method looks for the specified key in the current level of hierarchy
	 * only.
	 * 
	 * @param key
	 * @return
	 */
	final public Leaf locate(final K key) {
		return map().get(key);
	}

	/**
	 * Looks for specified object through the current item and all subitems.
	 * 
	 * @param key
	 * @return
	 */
	final public Leaf locateDeep(final K key) {
		Leaf node = locate(key);
		if (node != null)
			return node;
		for (Iterator<Leaf> i = map().values().iterator(); i.hasNext();) {
			Leaf subtree = i.next();
			node = subtree.locateDeep(key);
			if (node != null)
				return node;
		}
		return null;
	}

	final public Iterator<Leaf> iterator() {
		return (Iterator<Leaf>)map().values().iterator();
	}

	/** This method is extremely unefficient in MapTree,
	 * so programmer shouldn't use it in usual conditions,
	 * it's more or less to compatibility with general Tree.
	 */
	@SuppressWarnings("unchecked")
	public Leaf getNode(int index) {
		Leaf trees[] = (Leaf[])map().values().toArray();
		return trees[index];
	}	
	
	private BuiltinMap<K,Leaf> map() {
		if(map==null)
			map=new BuiltinMap<K,Leaf>();
		return map;
	}
	
	private K key;
	transient private BuiltinMap<K,Leaf> map;
	
	/*
	 * UNIT TEST
	 */

	@SuppressWarnings("unchecked")
	public static void main(String args[]) {
		class MT extends MapTree<String,MT> {
			private static final long serialVersionUID = -3974244097314747236L;

			public MT(String name) {
				super(name);
			}
		}

		MT root = new MT("root");
		MT r1 = new MT("1");
		MT r11 = new MT("1");
		MT r12 = new MT("2");
		MT r13 = new MT("3");
		MT r2 = new MT("2");
		MT r3 = new MT("3");
		MT r31 = new MT("31");
		MT r311 = new MT("311");
		MT r312 = new MT("312");
		root.add(r1);
		r1.add(r11);
		r1.add(r12);
		r1.add(r13);
		root.add(r2);
		
		SerializationTest.writeObject(root, MapTree.class.getName());

		root.add(r3);
		r3.add(r31);
		r31.add(r311);
		r31.add(r312);

		System.out.println(r312.format());

		MT t1 = (MT) r3.locate("1");
		MT t2 = (MT) r3.locateDeep("311");		
		MT t3 = (MT) r3.locate("31");
		
		System.out.println("format:"+t2.format(".", true));

		root.print(System.out);
		System.out.println("t1=" + t1);
		System.out.println("t2=" + t2.format());
		System.out.println("t3=" + t3.format());
		
		CompoundName<String> scn=new StringCompoundName("root.3.31.311");
		CompoundName<String> scn2=new StringCompoundName("3.31.311");

		System.out.println();
		System.out.println(root.locate(scn, true));
		System.out.println(root.locate(scn2, false));
		System.out.println(root.locate(scn, false));
		System.out.println(root.locate(scn2, true));
		
		
		System.out.println("*************");
		Tree t = (Tree) SerializationTest.readObject(MapTree.class.getName());
		t.print(System.out);
	}
}
