package com.ssv.utils.tree;

import java.io.IOException;
import java.io.PrintStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import com.ssv.utils.CompoundName;

abstract public class Tree<Leaf extends Tree<Leaf>> 
implements CompoundName<Tree<Leaf>>, Serializable {
	/**
	 * This constructs the root node.
	 */
	public Tree() {
	}

	public Tree<Leaf> getParent() {
		return parent;
	}

	final public Tree<Leaf> getLastToken() {
		return this;
	}

	final public String format(String divider, boolean showRoot) {
		StringBuffer sb = new StringBuffer();
		if (showRoot)
			getName(divider, sb);
		else
			this.getNameWithoutRoot(divider, sb);
		return sb.toString();
	}

	final public String format(boolean showRoot) {
		return format(DOT, showRoot);
	}

	final public String format() {
		return format(DOT, true);
	}

	final public Tree<Leaf> getToken(int index) {
		if (index == numTokens() - 1)
			return getLastToken();
		return getPrefix().getToken(index);
	}

	@SuppressWarnings("unchecked")
	final public Tree<Leaf>[] toArray() {
		Tree<Leaf> arr[] = (Tree<Leaf>[])new Object[numTokens()];
		for (int i = 0; i < numTokens(); i++)
			arr[i] = getToken(i);
		return arr;
	}

	final public CompoundName<Tree<Leaf>> getPrefix() {
		return getParent();
	}

	final public int numTokens() {
		if (getPrefix() == null)
			return 1;
		return getPrefix().numTokens() + 1;
	}

	public final boolean add(Leaf tree) {
		if (tree != null)
			tree.setParent(this);
		return addNode(tree);
	}

	public final boolean remove(Leaf tree) {
		boolean rez = removeNode(tree);
		tree.setParent(null);
		return rez;
	}

	@SuppressWarnings("unchecked")
	final private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		out.writeInt(size());
		for (Iterator i = iterator(); i.hasNext();) {
			out.writeObject(i.next());
		}
	}

	@SuppressWarnings("unchecked")
	final private void readObject(java.io.ObjectInputStream in) throws IOException,
			ClassNotFoundException {
		int sz = in.readInt();
		for (int i = 0; i < sz; i++) {
			add((Leaf) in.readObject());
		}
	}
	
	/**
	 * Appends the specified element to the tree.
	 * 
	 * @return true if this collection changed as a result of the call.
	 */
	abstract protected boolean addNode(Leaf tree);

	/**
	 * @return number of subnodes of current node.
	 */
	abstract public int size();

	/**
	 * Removes the element.
	 * 
	 * @return true if the tree contained the specified element.
	 */
	abstract protected boolean removeNode(Leaf tree);

	/**
	 * @param index
	 * @return the node which adressed by index.
	 */
	abstract public Leaf getNode(int index);

	/**
	 * Returns an iterator over the subnodes in this tree.
	 * 
	 * @return
	 */
	public Iterator<Leaf> iterator() {
		return new DefaultTreeIterator<Leaf>(this);
	}

	/**
	 * @return true if the node doesn't have subnodes, false otherwise.
	 */
	final public boolean isTerminal() {
		return size() == 0;
	}

	public void print(PrintStream ps) {
		printPrefix(ps);
	}

	public void printPrefix(PrintStream ps) {
		prefix(new PrintOperation<Leaf>(ps));
	}

	public void printPostfix(PrintStream ps) {
		postfix(new PrintOperation<Leaf>(ps));
	}

	final public Leaf prefix(ITreeAction<Leaf> action) {
		Leaf rez;
		if ((rez = action.performAction(this)) != null)
			return rez;
		action.enter();
		for (Iterator<Leaf> i = iterator(); i.hasNext();) {
			if ((rez = i.next().prefix(action)) != null)
				return rez;
		}
		action.exit();
		return null;
	}

	final public Leaf postfix(ITreeAction<Leaf> action) {
		Leaf rez;
		action.enter();
		for (Iterator<Leaf> i = iterator(); i.hasNext();) {
			if ((rez = i.next().postfix(action)) != null)
				return rez;
		}
		action.exit();
		if ((rez = action.performAction(this)) != null)
			return rez;
		return null;
	}
	
	final public Leaf prefix(ITreeAction<Leaf> action, 
			Comparator<Leaf> comparator) {
		Leaf rez;
		if ((rez = action.performAction(this)) != null)
			return rez;
		action.enter();

		List<Leaf> list = new ArrayList<Leaf>();

		for (Iterator<Leaf> i = iterator(); i.hasNext();) {
			list.add(i.next());
		}

		Collections.sort(list, comparator);

		for (Iterator<Leaf> i = list.iterator(); i.hasNext();) {
			if ((rez = i.next().prefix(action, comparator)) != null)
				return rez;
		}
		action.exit();
		return null;
	}

	final public Leaf postfix(ITreeAction<Leaf> action, 
			Comparator<Leaf> comparator) {
		Leaf rez;
		action.enter();

		List<Leaf> list = new ArrayList<Leaf>();

		for (Iterator<Leaf> i = iterator(); i.hasNext();) {
			list.add(i.next());
		}

		Collections.sort(list, comparator);

		for (Iterator<Leaf> i = list.iterator(); i.hasNext();) {
			if ((rez = i.next().postfix(action, comparator)) != null)
				return rez;
		}
		
		action.exit();
		if ((rez = action.performAction(this)) != null)
			return rez;
		return null;
	}

	/**
	 * This class should be used only to prefix printing whole hierarchy
	 * 
	 * @author SSiryk
	 */
	final static protected class PrintOperation<L extends Tree<L>> 
	extends SimplePrintOperation<L> {
		public PrintOperation(PrintStream ps) {
			super(ps);
		}

		void printItem(Tree<L> item) {
			ps.println(item);
		}
	}

	protected static abstract class SimplePrintOperation<L extends Tree<L>> 
	extends AbstractTreeAction<L> {
		public SimplePrintOperation(PrintStream ps) {
			this.ps = ps;
		}

		public L performAction(Tree<L> item) {
			printTab();
			printItem(item);
			return null;
		}

		abstract void printItem(Tree<L> item);

		private void printTab() {
			for (int i = 0; i < this.getLevel(); i++)
				ps.print("\t");
		}

		PrintStream ps;
	};

	private void getName(String divider, StringBuffer sb) {
		if (getPrefix() != null) {
			getParent().getName(divider, sb);
			sb.append(divider);
		}
		sb.append(getLastToken().toString());
	}

	private void getNameWithoutRoot(String divider, StringBuffer sb) {
		if (getPrefix() != null) {
			getParent().getNameWithoutRoot(divider, sb);
			if (getParent().getParent() != null)
				sb.append(divider);
			sb.append(getLastToken().toString());
		}
	}

	final protected void setParent(Tree<Leaf> parent) {
		this.parent = parent;
	}

	transient private Tree<Leaf> parent;

	/** @link dependency */
	/* # ITreeAction lnkITreeAction; */

	/* UNIT TEST */

	public static void main(String[] args) {
		System.out.println("\n\nDefaultTree");
		DefaultTree.main(args);
		System.out.println("\n\nMapTree");
		MapTree.main(args);
		System.out.println("\n\nBinaryTree");
		BinaryTree.main(args);
	}
}
